local XPerl_Player_Events = {}
local isOutOfControl = nil
local playerClass
local conf
XPerl_RequestConfig(function(new) conf = new end)
local perc1F = "%.1f%%"
local perc3D = "%3d%%"

local format = format
local GetNumRaidMembers = GetNumRaidMembers
local UnitHealth = UnitHealth
local UnitHealthMax = UnitHealthMax
local UnitIsConnected = UnitIsConnected
local UnitIsDead = UnitIsDead
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitIsGhost = UnitIsGhost
local UnitMana = UnitMana
local UnitManaMax = UnitManaMax
local UnitName = UnitName
local UnitPowerType = UnitPowerType

----------------------
-- Loading Function --
----------------------
function XPerl_Player_OnLoad(self)
	XPerl_SetChildMembers(self)
	self.partyid = "player"

	XPerl_BlizzFrameDisable(PlayerFrame)

	CombatFeedback_Initialize(self.hitIndicator.text, 30)

        XPerl_SecureUnitButton_OnLoad(self, "player", XPerl_ShowGenericMenu)
        XPerl_SecureUnitButton_OnLoad(self.nameFrame, "player", XPerl_ShowGenericMenu)

	local events = {"PLAYER_ENTERING_WORLD", "VARIABLES_LOADED", "PARTY_MEMBERS_CHANGED", "PARTY_LEADER_CHANGED",
			"PARTY_LOOT_METHOD_CHANGED", "RAID_ROSTER_UPDATE", "PLAYER_UPDATE_RESTING", "PLAYER_REGEN_ENABLED",
			"PLAYER_REGEN_DISABLED", "PLAYER_ENTER_COMBAT", "PLAYER_LEAVE_COMBAT", "PLAYER_LEVEL_UP", "PLAYER_DEAD",
			"PLAYER_XP_UPDATE", "UPDATE_FACTION", "PLAYER_AURAS_CHANGED", "PLAYER_CONTROL_LOST", "PLAYER_CONTROL_GAINED",
			"UNIT_COMBAT", "UPDATE_STEALTH"}

	for i,event in pairs(events) do
		self:RegisterEvent(event)
	end

	XPerl_UnitEvents(self, XPerl_Player_Events, {"UNIT_SPELLMISS", "UNIT_FACTION", "UNIT_PORTRAIT_UPDATE", "UNIT_FLAGS", "PLAYER_FLAGS_CHANGED"})
	XPerl_RegisterBasics(self, XPerl_Player_Events)

	self.statsFrame.druidBar:SetHeight(1)
	self.statsFrame.druidBar:Hide()

	self.EnergyTime = 0
	self.EnergyLast = 0

	self:SetScript("OnUpdate", XPerl_Player_OnUpdate)
	self:SetScript("OnEvent", XPerl_Player_OnEvent)
	self:SetScript("OnShow", XPerl_Player_UpdatePortrait)
	self.time = 0

	XPerl_RegisterHighlight(self.highlight, 3)
	XPerl_RegisterPerlFrames(self, {self.nameFrame, self.statsFrame, self.levelFrame, self.portraitFrame, self.groupFrame})

	if (XPerl_ArcaneBar_RegisterFrame) then
		XPerl_ArcaneBar_RegisterFrame(self.nameFrame, "player", "ArcaneBarPlayer")
	end

	self.FlashFrames = {self.portraitFrame, self.nameFrame,
				self.levelFrame, self.statsFrame}

	XPerl_RegisterOptionChanger(XPerl_Player_Set_Bits, self)
	XPerl_Highlight:Register(XPerl_Player_HighlightCallback)
	self.IgnoreHighlightStates = {AGGRO = true}
end

-- XPerl_Player_HighlightCallback(updateName)
function XPerl_Player_HighlightCallback(updateName)
	if (UnitName("player") == updateName) then
		XPerl_Highlight:SetHighlight(XPerl_Player, updateName)
	end
end

-------------------------
-- The Update Function --
-------------------------

local function XPerl_Player_CombatFlash(self, elapsed, argNew, argGreen)
	if (XPerl_CombatFlashSet (elapsed, self, argNew, argGreen)) then
		XPerl_CombatFlashSetFrames(self)
	end
end

-- XPerl_Player_UpdatePortrait
function XPerl_Player_UpdatePortrait(self)
	if (conf and conf.ShowPlayerPortrait == 1) then
		if (conf.ShowPlayerPortrait3D == 1) then
			self.portraitFrame.portrait:Hide()
			self.portraitFrame.portrait3D:Show()
			XPerlSetPortrait3D(self.portraitFrame.portrait3D, "player")
		else
			self.portraitFrame.portrait3D.last3DTime = nil
			self.portraitFrame.portrait:Show()
			self.portraitFrame.portrait3D:Hide()
			SetPortraitTexture(self.portraitFrame.portrait, "player")
		end
	end
end

-- XPerl_Player_UpdateManaType
local function XPerl_Player_UpdateManaType(self)
	XPerl_SetManaBarType(self)

	if (UnitPowerType("player") == 3) then
		self.EnergyEnabled = true
	else
		self.EnergyEnabled = nil
	end
	XPerl_Player_TickerShowHide(self)
end

-- XPerl_Player_UpdateLeader()
function XPerl_Player_UpdateLeader(self)
	if (IsPartyLeader()) then
		self.nameFrame.leaderIcon:Show()
	else
		self.nameFrame.leaderIcon:Hide()
	end

	if (conf.ShowPartyNumber == 1 and GetNumRaidMembers() > 0) then
		for i = 1,GetNumRaidMembers() do
			local name, rank, subgroup = GetRaidRosterInfo(i)
			if (name == UnitName("player")) then
				if (conf.ShowPartyNumberWithName == 0) then
					self.groupFrame.text:SetText(format(XPERL_RAID_GROUP, subgroup))
					self.groupFrame:Show()
					self.nameFrame.group:Hide()
					return
				else
					self.nameFrame.group:SetText(format(XPERL_RAID_GROUPSHORT, subgroup))
					self.nameFrame.group:Show()
					self.groupFrame:Hide()
					return
				end
			end
		end
	end

	self.nameFrame.group:Hide()
	self.groupFrame:Hide()
end

-- XPerl_Player_UpdateMasterLooter()
local function XPerl_Player_UpdateMasterLooter(self)
	local method, index

	if (UnitInParty("party1") or UnitInRaid("player")) then
		method, index = GetLootMethod()
	end

	if (method == "master" and index == 0) then
		self.nameFrame.masterIcon:Show()
	else
		self.nameFrame.masterIcon:Hide()
	end
end

-- XPerl_Player_UpdateName()
local function XPerl_Player_UpdateName(self)
	self.nameFrame.text:SetText(UnitName("player"))
end

-- XPerl_Player_UpdateRep
local function XPerl_Player_UpdateRep(self)
	local name, reaction, min, max, value = GetWatchedFactionInfo()
	local color

	if (name) then
		color = FACTION_BAR_COLORS[reaction]
	else
		name = XPERL_LOC_NONEWATCHED
		max = 1
		min = 0
		value = 0
		color = FACTION_BAR_COLORS[4]
	end

	value = value - min
	max = max - min
	min = 0

	self.statsFrame.repBar:SetMinMaxValues(min, max)
	self.statsFrame.repBar:SetValue(value)

	self.statsFrame.repBar:SetStatusBarColor(color.r, color.g, color.b, 1)
	self.statsFrame.repBar.bg:SetVertexColor(color.r, color.g, color.b, 0.25)

	local perc = (value * 100) / max
	self.statsFrame.repBar.percent:SetText(format(perc1F, perc))

	if (max == 1) then
		self.statsFrame.repBar.text:SetText(name)
	else
		self.statsFrame.repBar.text:SetText(format("%d/%d", value, max))
	end
	--self.statsFrame.repBar.text:SetText(name)
end

-- XPerl_Player_UpdateXP
local function XPerl_Player_UpdateXP(self)
	local playerxp = UnitXP("player")
	local playerxpmax = UnitXPMax("player")
	local playerxprest = GetXPExhaustion() or 0
	self.statsFrame.xpBar:SetMinMaxValues(0, playerxpmax)
	self.statsFrame.xpRestBar:SetMinMaxValues(0, playerxpmax)
	self.statsFrame.xpBar:SetValue(playerxp)
	local xptext

	if (playerxpmax > 10000) then
		xptext = format("%dK/%dK", playerxp / 1000, playerxpmax / 1000)
	else
		xptext = format("%d/%d", playerxp, playerxpmax)
	end

	if (playerxprest > 0) then
		if (playerxpmax > 10000) then
			xptext = xptext .. format("(+%dK)", playerxprest / 1000)
		else
			xptext = xptext .. format("(+%d)", playerxprest)
		end

		color = {r = 0.3, g = 0.3, b = 1}
	else
		color = {r = 0.6, g = 0, b = 0.6}
	end

	self.statsFrame.xpRestBar:SetValue(playerxp + playerxprest)
	self.statsFrame.xpBar:SetStatusBarColor(color.r, color.g, color.b, 1)
	self.statsFrame.xpRestBar:SetStatusBarColor(color.r, color.g, color.b, 0.5)
	self.statsFrame.xpBar.bg:SetVertexColor(color.r, color.g, color.b, 0.25)

	self.statsFrame.xpBar.text:SetText(xptext)
	self.statsFrame.xpBar.percent:SetText(format(perc3D, (playerxp * 100.0) / playerxpmax))
end

-- XPerl_Player_UpdateCombat
local function XPerl_Player_UpdateCombat(self)
	if (UnitAffectingCombat("player")) then
		self.nameFrame.text:SetTextColor(1,0,0)
		self.nameFrame.combatIcon:SetTexCoord(0.5, 1.0, 0.0, 0.5)
		self.nameFrame.combatIcon:Show()
	else
		if (UnitIsPVP("player")) then
			self.nameFrame.text:SetTextColor(0,1,0)
		else
			XPerl_ColourFriendlyUnit(self.nameFrame.text, "player")
		end

		if (IsResting()) then
			self.nameFrame.combatIcon:SetTexCoord(0, 0.5, 0.0, 0.5)
			self.nameFrame.combatIcon:Show()
		else
			self.nameFrame.combatIcon:Hide()
		end
	end
end

-- XPerl_Player_UpdatePVP
local function XPerl_Player_UpdatePVP(self)
	--local playerrankname, playerrank=GetPVPRankInfo(UnitPVPRank("player"), "player")
	--if (playerrank and conf.ShowPlayerPVPRank==1) then
	--	self.nameFrame.pvpRankIcon:Show()
	--	if (playerrank==0) then
	--		self.nameFrame.pvpRankIcon:Hide()
	--	else
	--		self.nameFrame.pvpRankIcon:SetTexture(format("Interface\\PVPRankBadges\\PVPRank%02d", playerrank))
	--	end
	--else
	--	self.nameFrame.pvpRankIcon:Hide()
	--end

	-- PVP Status settings
	if (UnitAffectingCombat("player")) then
		self.nameFrame.text:SetTextColor(1,0,0)
	elseif (conf.ClassColouredNames == 0 and UnitIsPVP("player")) then
		self.nameFrame.text:SetTextColor(0,1,0)
	else
		XPerl_ColourFriendlyUnit(self.nameFrame.text, "player")
	end

	if (conf.ShowPlayerPVP == 1 and UnitIsPVP("player")) then
		self.nameFrame.pvpIcon:Show()
		self.nameFrame.pvpIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-"..UnitFactionGroup("player"))
	else
		self.nameFrame.pvpIcon:Hide()
	end
end

-- XPerl_Player_Feigning
local function XPerl_Player_Feigning (self)
	for buffnum=1,16 do
		local buff = UnitBuff("player", buffnum)
		if (not buff) then
			return
		end
		if (strfind(strlower(buff), "feigndeath")) then
			return true
		end
	end
end

-- XPerl_Player_DruidBarUpdate
local function XPerl_Player_DruidBarUpdate(self)
	local manaPct
	if (SimpleDruidBar) then
		self.statsFrame.druidBar:SetMinMaxValues(0, SimpleDruidBar.MaxMana)
		self.statsFrame.druidBar:SetValue(SimpleDruidBar.CurrMana)
		manaPct = (SimpleDruidBar.CurrMana * 100.0) / SimpleDruidBar.MaxMana
		self.statsFrame.druidBar.text:SetText(format("%d/%d", ceil(SimpleDruidBar.CurrMana), SimpleDruidBar.MaxMana))

		-- Doesn't hide the frame, but makes it not appear on screen. The code relies on IsShown() to function
		SimpleDruidBar.DISPLAY_BAR:ClearAllPoints()
	else
		self.statsFrame.druidBar:SetMinMaxValues(0,DruidBarKey.maxmana)
		self.statsFrame.druidBar:SetValue(DruidBarKey.keepthemana)
		manaPct = (DruidBarKey.keepthemana * 100.0) / DruidBarKey.maxmana
		self.statsFrame.druidBar.text:SetText(format("%d/%d", DruidBarKey.keepthemana, DruidBarKey.maxmana))
	end

	self.statsFrame.druidBar.percent:SetText(format(perc3D, manaPct))

	local druidBarExtra
	if (UnitPowerType("player")>0) then
		self.statsFrame.druidBar.text:Show()
		if (conf.ShowPlayerPercent == 1) then
			self.statsFrame.druidBar.percent:Show()
		end
		self.statsFrame.druidBar:Show()
		self.statsFrame.druidBar:SetHeight(10)
		druidBarExtra = 1
	else
		self.statsFrame.druidBar.percent:Hide()
		self.statsFrame.druidBar.text:Hide()
		self.statsFrame.druidBar:Hide()
		self.statsFrame.druidBar:SetHeight(1)
		druidBarExtra = 0
	end

	local h = 40 + ((druidBarExtra + conf.ShowPlayerRepBar + conf.ShowPlayerXPBar) * 10)
	XPerl_Player.statsFrame:SetHeight(h)
	XPerl_StatsFrameSetup(self, {"druidBar", "xpBar", "repBar"})
end

-- XPerl_Player_UpdateMana
local function XPerl_Player_UpdateMana(self)
	local playermana = UnitMana("player")
	local playermanamax = UnitManaMax("player")
	self.statsFrame.manaBar:SetMinMaxValues(0, playermanamax)
	self.statsFrame.manaBar:SetValue(playermana)

	self.statsFrame.manaBar.text:SetText(format("%d/%d", playermana, playermanamax))
	if (UnitPowerType("player") >= 1) then
		self.statsFrame.manaBar.percent:SetText(playermana)
	else
		self.statsFrame.manaBar.percent:SetText(format(perc3D, (playermana * 100.0) / playermanamax))
	end

	if (conf.ShowPlayerValues == 1) then
		self.statsFrame.manaBar.text:Show()
	else
		self.statsFrame.manaBar.text:Hide()
	end

	if (DruidBarKey or SimpleDruidBar) then
		if (playerClass == "DRUID") then
			XPerl_Player_DruidBarUpdate(self)
		end
	end
end

-- XPerl_Player_UpdateHealth
local function XPerl_Player_UpdateHealth(self)
	local playerhealth = UnitHealth("player")
	local playerhealthmax = UnitHealthMax("player")

	if (playerhealth > playerhealthmax) then
		-- New glitch with 1.12.1
		if (UnitIsDeadOrGhost("player")) then
			playerhealth = 0
		else
			playerhealth = playerhealthmax
		end
	end

	XPerl_SetHealthBar(self.statsFrame.healthBar, playerhealth, playerhealthmax, conf.HealerModePlayer, conf.HealerModePlayerType)

	local greyMsg
	if (UnitIsDead("player")) then
		if (XPerl_Player_Feigning()) then
			greyMsg = XPERL_LOC_FEIGNDEATH
		else
			greyMsg = XPERL_LOC_DEAD
			self.statsFrame.manaBar.text:Hide()
		end

	elseif (UnitIsGhost("player")) then
		greyMsg = XPERL_LOC_GHOST
		self.statsFrame.manaBar.text:Hide()

	elseif (UnitIsAFK("player") and conf.ShowAFK ~= 0) then
		greyMsg = CHAT_MSG_AFK
	end

	if (greyMsg) then
		if (conf.ShowPlayerPercent == 1) then
			self.statsFrame.healthBar.percent:SetText(greyMsg)
			self.statsFrame.healthBar.percent:Show()
		else
			self.statsFrame.healthBar.text:SetText(greyMsg)
			self.statsFrame.healthBar.text:Show()
		end

		self.statsFrame:SetGrey()
	else
		if (self.statsFrame.greyMana) then
			if (conf.ShowPlayerValues == 0) then
				self.statsFrame.healthBar.text:Hide()
				self.statsFrame.manaBar.text:Hide()
			end

			self.statsFrame.greyMana = nil
			XPerl_Player_UpdateManaType(self)
		end
	end

	XPerl_PlayerStatus_OnUpdate(playerhealth, playerhealthmax)
end

-- XPerl_Player_UpdateLevel
local function XPerl_Player_UpdateLevel(self)
	self.levelFrame.text:SetText(UnitLevel("player"))
end

-- XPerl_PlayerStatus_OnUpdate
function XPerl_PlayerStatus_OnUpdate(val, max)
	if (conf.FullScreenStatus == 1) then
		local testLow = conf.FullScreenStatusWarn / 100
		local testHigh = conf.FullScreenStatusOK / 100

		if (val and max) then
			local test = val / max

			if ( test <= testLow and not XPerl_LowHealthFrame.frameFlash and not UnitIsDeadOrGhost("player")) then
				XPerl_FrameFlash(XPerl_LowHealthFrame)
			elseif ( (test >= testHigh and XPerl_LowHealthFrame.frameFlash) or UnitIsDeadOrGhost("player") ) then
				XPerl_FrameFlashStop(XPerl_LowHealthFrame, "out")
			end
		else
			if (isOutOfControl and not XPerl_OutOfControlFrame.frameFlash and not UnitOnTaxi("player")) then
				XPerl_FrameFlash(XPerl_OutOfControlFrame)
			elseif (not isOutOfControl and XPerl_OutOfControlFrame.frameFlash) then
				XPerl_FrameFlashStop(XPerl_OutOfControlFrame, "out")
			end
		end
	else
		if (XPerl_LowHealthFrame.frameFlash) then
			XPerl_FrameFlashStop(XPerl_LowHealthFrame)
		end
		if (XPerl_OutOfControlFrame.frameFlash) then
			XPerl_FrameFlashStop(XPerl_OutOfControlFrame)
		end
	end
end

-- XPerl_Player_OnUpdate
function XPerl_Player_OnUpdate(self, elapsed)
	CombatFeedback_OnUpdate(elapsed)
	if (self.PlayerFlash) then
		XPerl_Player_CombatFlash(self, elapsed, false)
	end
end

-- XPerl_Player_UpdateBuffs
local function XPerl_Player_UpdateBuffs(self)
	XPerl_CheckDebuffs(self, "player")

	if (DruidBarKey or SimpleDruidBar) then
		-- For DruidBar addon, we update the mana bar on shapeshift
		if (playerClass == "DRUID") then
			XPerl_Player_UpdateMana(self)
		end
	end

	if (conf.FullScreenStatus == 1) then
		if (isOutOfControl and not UnitOnTaxi("player")) then
			XPerl_PlayerStatus_OnUpdate()
		end
	end
end

-- XPerl_Player_UpdateDisplay
function XPerl_Player_UpdateDisplay (self)
	XPerl_Player_UpdateXP(self)
	XPerl_Player_UpdateRep(self)
	XPerl_Player_UpdateManaType(self)
	XPerl_Player_UpdateLevel(self)
	XPerl_Player_UpdateName(self)
	XPerl_Player_UpdatePVP(self)
	XPerl_Player_UpdateCombat(self)
	XPerl_Player_UpdateLeader(self)
	XPerl_Player_UpdateMasterLooter(self)
	XPerl_Player_UpdateMana(self)
	XPerl_Player_UpdateHealth(self)
	XPerl_Player_UpdateBuffs(self)
end

-- EVENTS AND STUFF

-------------------
-- Event Handler --
-------------------
function XPerl_Player_OnEvent(self, event, a, b)
	local func = XPerl_Player_Events[event]
	if (func) then
		func(self, a, b)
	else
XPerl_ShowMessage("EXTRA EVENT")
	end
end

-- PLAYER_ENTERING_WORLD
function XPerl_Player_Events:PLAYER_ENTERING_WORLD()
	XPerl_Player_UpdateDisplay(self)
end

-- UNIT_COMBAT
function XPerl_Player_Events:UNIT_COMBAT()
 	if (arg1 == "player") then
		if (conf.CombatHitIndicator == 1 and conf.ShowPlayerPortrait == 1) then
			CombatFeedback_OnCombatEvent(arg2, arg3, arg4, arg5)
		end

		if (arg2 == "HEAL") then
			XPerl_Player_CombatFlash(self, elapsed, true, true)
		elseif (arg4 and arg4 > 0) then
			XPerl_Player_CombatFlash(self, elapsed, true)
		end
	end
end

-- UNIT_SPELLMISS
function XPerl_Player_Events:UNIT_SPELLMISS()
	if (arg1 == "player") then
		if (conf.CombatHitIndicator == 1 and conf.ShowPlayerPortrait == 1) then
			CombatFeedback_OnSpellMissEvent(arg2)
		end
	end
end

-- UNIT_PORTRAIT_UPDATE
function XPerl_Player_Events:UNIT_PORTRAIT_UPDATE()
	if (arg1 == "player") then
		XPerl_Player_UpdatePortrait(self)
	end
end

-- VARIABLES_LOADED
function XPerl_Player_Events:VARIABLES_LOADED()
	perc1F = "%.1f"..PERCENT_SYMBOL
	perc3D = "%3d"..PERCENT_SYMBOL
	self.doubleCheckAFK = 2		-- Check during 2nd UPDATE_FACTION, which are the last guarenteed events to come after logging in
	XPerl_Player_UpdatePortrait(self)
	self:UnregisterEvent(event)
end

-- PARTY_LOOT_METHOD_CHANGED
function XPerl_Player_Events:PARTY_LOOT_METHOD_CHANGED()
	XPerl_Player_UpdateLeader(self)
	XPerl_Player_UpdateMasterLooter(self)
end
XPerl_Player_Events.PARTY_MEMBERS_CHANGED = XPerl_Player_Events.PARTY_LOOT_METHOD_CHANGED
XPerl_Player_Events.PARTY_LEADER_CHANGED  = XPerl_Player_Events.PARTY_LOOT_METHOD_CHANGED

function XPerl_Player_Events:RAID_ROSTER_UPDATE()
	XPerl_Player_UpdateLeader(self)
	XPerl_Player_UpdateMasterLooter(self)
end

-- UNIT_HEALTH, UNIT_MAXHEALTH
function XPerl_Player_Events:UNIT_HEALTH()
	XPerl_Player_UpdateHealth(self)
end
XPerl_Player_Events.UNIT_MAXHEALTH = XPerl_Player_Events.UNIT_HEALTH
XPerl_Player_Events.PLAYER_DEAD    = XPerl_Player_Events.UNIT_HEALTH

-- UNIT_MANA, UNIT_MAXMANA, UNIT_RAGE, UNIT_MAXRAGE, UNIT_ENERGY
-- UNIT_MAXENERGY
function XPerl_Player_Events:UNIT_MANA()
	XPerl_Player_UpdateMana(self)
end
XPerl_Player_Events.UNIT_MAXMANA		= XPerl_Player_Events.UNIT_MANA
XPerl_Player_Events.UNIT_RAGE		= XPerl_Player_Events.UNIT_MANA
XPerl_Player_Events.UNIT_MAXRAGE         = XPerl_Player_Events.UNIT_MANA

function XPerl_Player_Events:UNIT_ENERGY()
	XPerl_Player_UpdateMana(self)

	local e = UnitMana("player")
	local m = UnitManaMax("player")
	if (e == self.EnergyLast + 20) then
		XPerl_Player.EnergyTime = GetTime()
	end
	self.EnergyLast = e

	XPerl_Player_TickerShowHide(self)
end
XPerl_Player_Events.UNIT_MAXENERGY       = XPerl_Player_Events.UNIT_ENERGY

-- UNIT_DISPLAYPOWER
function XPerl_Player_Events:UNIT_DISPLAYPOWER()
	XPerl_Player_UpdateManaType(self)
	XPerl_Player_UpdateMana(self)
	XPerl_Player_TickerShowHide(self)
end

-- UNIT_NAME_UPDATE
function XPerl_Player_Events:UNIT_NAME_UPDATE()
	XPerl_Player_UpdateName(self)
end

-- UNIT_LEVEL
function XPerl_Player_Events:UNIT_LEVEL()
	XPerl_Player_UpdateLevel(self)
	XPerl_Player_UpdateXP(self)
end
XPerl_Player_Events.PLAYER_LEVEL_UP = XPerl_Player_Events.UNIT_LEVEL

-- PLAYER_XP_UPDATE
function XPerl_Player_Events:PLAYER_XP_UPDATE()
	XPerl_Player_UpdateXP(self)
end

-- UPDATE_STEALTH
function XPerl_Player_Events:UPDATE_STEALTH()
	XPerl_Player_TickerShowHide(self)
end

-- UPDATE_FACTION
function XPerl_Player_Events:UPDATE_FACTION()
	XPerl_Player_UpdateRep(self)

	if (self.doubleCheckAFK) then
		self.doubleCheckAFK = self.doubleCheckAFK - 1
		if (self.doubleCheckAFK <= 0) then
			XPerl_Player_UpdateHealth(self)
			self.doubleCheckAFK = nil
		end
	end
end

-- UNIT_FACTION
function XPerl_Player_Events:UNIT_FACTION()
	XPerl_Player_UpdatePVP(self)
end
XPerl_Player_Events.UNIT_FLAGS = XPerl_Player_Events.UNIT_FACTION

function XPerl_Player_Events:PLAYER_FLAGS_CHANGED(unit)
	XPerl_Player_UpdateHealth(self)
end

-- PLAYER_ENTER_COMBAT, PLAYER_LEAVE_COMBAT
function XPerl_Player_Events:PLAYER_ENTER_COMBAT()
	XPerl_Player_UpdateCombat(self)
end
XPerl_Player_Events.PLAYER_LEAVE_COMBAT = XPerl_Player_Events.PLAYER_ENTER_COMBAT

-- PLAYER_REGEN_ENABLED, PLAYER_REGEN_DISABLED
function XPerl_Player_Events:PLAYER_REGEN_ENABLED()
	XPerl_Player_UpdateCombat(self)
end
XPerl_Player_Events.PLAYER_REGEN_DISABLED = XPerl_Player_Events.PLAYER_REGEN_ENABLED
XPerl_Player_Events.PLAYER_UPDATE_RESTING = XPerl_Player_Events.PLAYER_REGEN_ENABLED

function XPerl_Player_Events:PLAYER_AURAS_CHANGED()
	XPerl_Player_UpdateBuffs(self)
end

-- PLAYER_CONTROL_LOST
function XPerl_Player_Events:PLAYER_CONTROL_LOST()
	if (conf.FullScreenStatus == 1 and not UnitOnTaxi("player")) then
		isOutOfControl = 1
	end
end

-- PLAYER_CONTROL_GAINED
function XPerl_Player_Events:PLAYER_CONTROL_GAINED()
	isOutOfControl = nil
	if (conf.FullScreenStatus == 1) then
		XPerl_PlayerStatus_OnUpdate()
	end
end

-- Moving stuff
function XPerl_Player_GetGap()
	if (XPerl_Target) then
		local playerLeft = XPerl_Player.statsFrame:GetRight() * XPerl_Player:GetEffectiveScale()
		local targetLeft = XPerl_Target:GetLeft() * XPerl_Target:GetEffectiveScale()
		local a = targetLeft - playerLeft
		return floor(floor((a + 0.01) * 100) / 100 + 4)
	end
	return 0
end

-- XPerl_Player_SetGap
function XPerl_Player_SetGap(newGap)
	if (type(newGap) == "number") then
		newGap = newGap - 4

		local function SetChildGap(self, other)
			if (self and other) then
				local top = other:GetTop()
				local left

				if (self == XPerl_Player) then
					left = ((self.statsFrame:GetRight() * self:GetEffectiveScale()) + newGap) / other:GetEffectiveScale()
				else
					left = ((self:GetRight() * self:GetEffectiveScale()) + newGap) / other:GetEffectiveScale()
				end

				if (self == XPerl_Target or self == XPerl_TargetTarget or self == XPerl_Focus) then
					if (self.levelFrame and self.levelFrame:IsShown()) then
						left = left + self.levelFrame:GetWidth()
					end
				end

				other:ClearAllPoints()
				other:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", left, top)
				other:SetUserPlaced(true)
			end
		end

		SetChildGap(XPerl_Player, XPerl_Target)
		SetChildGap(XPerl_Target, XPerl_TargetTarget)
		SetChildGap(XPerl_TargetTarget, XPerl_TargetTargetTarget)
		SetChildGap(XPerl_Focus, XPerl_FocusTarget)
		SetChildGap(XPerl_Player_Pet, XPerl_PetTarget)
	end
end

-- XPerl_Player_AlignTop
function XPerl_Player_AlignTop()

	-- We set this for 1 reason, so that all the related frames scale in the same direction should the user do that...
	XPerl_Player:ClearAllPoints()
	XPerl_Player:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", XPerl_Player:GetLeft() or 5, XPerl_Player:GetTop() or 948)
	XPerl_Player:SetUserPlaced(true)

	local function AlignChildTop(self, other)
		if (self and other) then
			local top = self:GetTop() * self:GetEffectiveScale()
			local otherLeft = other:GetLeft() or 220
			local selfLeft = self:GetLeft() or 5

			if (otherLeft == nil) then
				otherLeft = selfLeft + 200
			end

			other:ClearAllPoints()
			local newTop = top / other:GetEffectiveScale()
			other:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", otherLeft, newTop)
			other:SetUserPlaced(true)
		end
	end

	AlignChildTop(XPerl_Player, XPerl_Target)
	AlignChildTop(XPerl_Target, XPerl_TargetTarget)
	AlignChildTop(XPerl_TargetTarget, XPerl_TargetTargetTarget)
	AlignChildTop(XPerl_Focus, XPerl_FocusTarget)
	AlignChildTop(XPerl_Player_Pet, XPerl_PetTarget)
end

-- XPerl_Player_Energy_TickWatch
function XPerl_Player_Energy_OnUpdate(self)

	local time = GetTime()
	local temp = time - XPerl_Player.EnergyTime
	local remainder = mod (temp, 2)

	self:SetValue(remainder)

	sparkPosition = self:GetWidth() * (remainder / 2)

	self.spark:SetPoint("CENTER", self, "LEFT", sparkPosition, 0)
end

-- XPerl_Player_TickerShowHide
function XPerl_Player_TickerShowHide(self)
       	local e = UnitMana("player")
       	local m = UnitManaMax("player")
       	if (self.EnergyEnabled and (e < m or UnitAffectingCombat("player") or IsStealthed()) and not UnitIsDeadOrGhost("player") and conf.EnergyTicker == 1) then
       		self.statsFrame.energyTicker:Show()
       	else
       		self.statsFrame.energyTicker:Hide()
       	end
end

-- XPerl_Player_SetWidth
function XPerl_Player_SetWidth(self)

	conf.PlayerWidthBonus = max(0, conf.PlayerWidthBonus or 0)
	if (conf.ShowPlayerPercent == 0) then
		self.nameFrame:SetWidth(128 + conf.PlayerWidthBonus)
		self.statsFrame:SetWidth(128 + conf.PlayerWidthBonus)
		self.statsFrame.healthBar.percent:Hide()
		self.statsFrame.manaBar.percent:Hide()
		self.statsFrame.xpBar.percent:Hide()
	else
		self.nameFrame:SetWidth(160 + conf.PlayerWidthBonus)
		self.statsFrame:SetWidth(160 + conf.PlayerWidthBonus)
		self.statsFrame.healthBar.percent:Show()
		self.statsFrame.manaBar.percent:Show()
		self.statsFrame.xpBar.percent:Show()
	end

	local h = 40 + (((self.statsFrame.druidBar:IsShown() or 0) + conf.ShowPlayerRepBar + conf.ShowPlayerXPBar) * 10)
	self.statsFrame:SetHeight(h)

	self:SetWidth(conf.ShowPlayerPortrait * 60 + 158 + conf.PlayerWidthBonus)
	self:SetScale(conf.Scale_PlayerFrame)

	XPerl_StatsFrameSetup(self, {"druidBar", "xpBar", "repBar"})
end

-- XPerl_Player_Set_Bits()
function XPerl_Player_Set_Bits(self)
	if (conf.ShowPlayerPortrait==0) then
		self.portraitFrame:Hide()
		self.portraitFrame:SetWidth(3)
	else
		self.portraitFrame:Show()
		self.portraitFrame:SetWidth(60)
		XPerl_Player_UpdatePortrait(self)
	end

	if (conf.ShowPlayerLevel==0) then
		self.levelFrame:Hide()
	else
		self.levelFrame:Show()
	end

	if (conf.ShowPlayerValues == 0) then
		self.statsFrame.healthBar.text:Hide()
		self.statsFrame.manaBar.text:Hide()
		self.statsFrame.xpBar.text:Hide()
	else
		self.statsFrame.healthBar.text:Show()
		self.statsFrame.manaBar.text:Show()
		self.statsFrame.xpBar.text:Show()
	end

	local _
	_, playerClass = UnitClass("player")
	local r, l, t, b = XPerl_ClassPos(playerClass)
	self.classFrame.tex:SetTexCoord(r, l, t, b)

	if (conf.ShowPlayerClassIcon == 0) then
		self.classFrame:Hide()
	else
		self.classFrame:Show()
	end

	if (conf.ShowPlayerRepBar == 0) then
		self.statsFrame.repBar:Hide()
	else
		self.statsFrame.repBar:Show()
	end

	if (conf.ShowPlayerXPBar == 0) then
		self.statsFrame.xpBar:Hide()
		self.statsFrame.xpRestBar:Hide()
	else
		self.statsFrame.xpBar:Show()
		self.statsFrame.xpRestBar:Show()
	end

	XPerl_Player_SetWidth(self)

	self:SetAlpha(conf.Transparency)

	XPerl_Player_UpdateDisplay(self)
end
